package rabbit.io;

// $Id: NLSOHandler.java,v 1.4 2005/08/03 17:00:51 robo Exp $

import java.io.IOException;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import rabbit.util.Logger;

/** A class to handle a group of NLSOInputStream's
 */
public class NLSOHandler {
    private Logger logger;

    private Selector selector;
    private List<NLSOInputStream> waitingChannels = 
    new ArrayList<NLSOInputStream> ();
    private Thread selectorThread;

    public NLSOHandler (Logger logger) throws IOException {
	this.logger = logger;
	setupSelector ();
    }

    public void addWaiting (NLSOInputStream nlso) {
	synchronized (waitingChannels) {
	    waitingChannels.add (nlso);
	}
	selector.wakeup ();
    }
    
    private synchronized void setupSelector () throws IOException {
	selector = Selector.open ();
	selectorThread = new Thread (new SelectRunnable (), 
				     "NLSOInputStreamHandler");
	selectorThread.start ();
    }

    private class SelectRunnable implements Runnable {
	public void run () {
	    while (true) {
		try {
		    int num = selector.select (5 * 1000); 
		    synchronized (waitingChannels) {
			if (selector.isOpen ()) {
			    handleSelects ();
			    handleReturned ();
			    cancelTimeouts ();
			}
		    }
		} catch (IOException e) {
		    logger.logError (Logger.ERROR, 
				     "NLSO: select failed: " + e);
		}
	    }
	}
    }

    private void handleSelects () throws IOException {
	Set<SelectionKey> selected = selector.selectedKeys ();
	for (Iterator<SelectionKey> i = selected.iterator (); i.hasNext (); ) {
	    SelectionKey sk = i.next ();
	    i.remove ();
	    NLSOInputStream nl = (NLSOInputStream)sk.attachment ();
	    if (sk.isConnectable ()) {
		nl.getSC ().finishConnect ();
		sk.cancel ();
		synchronized (nl) {
		    nl.setWaiting (false);
		    nl.notifyAll ();
		}
	    } else if (sk.isReadable ()) {
		sk.cancel ();
		synchronized (nl) {
		    nl.setReady ();
		    nl.setWaiting (false);
		    nl.notifyAll ();
		}
	    }
	}
    }
    
    private void handleReturned () throws IOException {
	try {
	    Long now = System.currentTimeMillis ();
	    for (NLSOInputStream nl : waitingChannels) {
		addChannel (now, nl);
	    }
	} finally {
	    waitingChannels.clear ();
	}
    }

    private void addChannel (long now, NLSOInputStream nl) 
	throws IOException {
	nl.setWaitTime (now);
	try {
	    SocketChannel sc = nl.getSC ();
	    sc.configureBlocking (false); 
	    if (sc.isConnectionPending () || !sc.isConnected ()) {
		sc.register (selector, SelectionKey.OP_CONNECT, nl);			
	    } else {
		sc.register (selector, SelectionKey.OP_READ, nl);
	    }
	} catch (CancelledKeyException cke) {
	    cke.printStackTrace ();
	    logger.logError (Logger.ALL, 
			     "NLSO: Returned channel cancelled, ignoring: " + 
			     cke);
	    try {
		nl.getSC ().close ();
	    } catch (IOException e) {
		logger.logError (Logger.ALL, 
				 "NLSO: Failed to close client socket: " + e);			
	    }
	} catch (ClosedChannelException cce) {
	    logger.logError (Logger.ALL, 
			     "NLSO: Closed channel: ignoring: " + cce);
	}
    }

    private void cancelTimeouts () throws IOException {
	long now = System.currentTimeMillis ();
	for (SelectionKey sk : selector.keys ()) {
	    NLSOInputStream nl = (NLSOInputStream)sk.attachment ();
	    if (now - nl.getWaitTime () > 25 * 1000) {
		sk.cancel ();
		try {
		    SocketChannel sc = (SocketChannel)sk.channel ();
		    sc.socket ().shutdownInput ();
		    sc.socket ().shutdownOutput ();		    
		    sc.close ();
		} catch (IOException e) {
		    logger.logError (Logger.ERROR, "NLSO: failed to shutdown and close socket: " + e);
		} finally {
		    synchronized (nl) {
			nl.setWaiting (false);
			nl.notifyAll ();
		    }
		}
	    }
	}
    }
}

