package rabbit.proxy;

import java.util.Date;
import java.util.List;
import java.util.StringTokenizer;
import rabbit.cache.NCacheEntry;
import rabbit.http.HTTPHeader;
import rabbit.util.Logger;

/** A class used to check for conditional requests. 
 */
class ConditionalChecker {
    
    private Logger logger;

    public ConditionalChecker (Logger logger) {
	this.logger = logger;
    }

    protected boolean checkConditional (Connection con, HTTPHeader header, 
					Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;	
	return (checkVary (con, header, rh) ||
		checkMaxAge (con, header, rh) || 
		checkNoCache (con, header, rh) ||
		checkQuery (con, header, rh) ||
		checkMinFresh (con, header, rh) || 
		checkRevalidation (con, header, rh));
    }

    private boolean checkVary (Connection con, HTTPHeader req, 
			       Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	if (entry == null)
	    return false;
	HTTPHeader resp = rh.getDataHook ();
	List varies = resp.getHeaders ("Vary");
	for (int i = 0; i < varies.size (); i++) {
	    String vary = (String)varies.get (i);
	    if (vary.equals ("*")) {
		con.setMayUseCache (false);
		return false;
	    } else {
		HTTPHeader origreq = (HTTPHeader)entry.getKey ();
		List vals = origreq.getHeaders (vary);
		List nvals = req.getHeaders (vary);
		if (vals.size () != nvals.size ()) {
		    return setupRevalidation (con, req, rh);
		} else {
		    for (int j = 0; j < vals.size (); j++) {
			String val = (String)vals.get (j);
			int k = nvals.indexOf (val);
			if (k == -1) {
			    return setupRevalidation (con, req, rh);
			}
		    }
		}
	    }
	}
	return false;
    }

    private boolean checkMaxAge (Connection con, HTTPHeader req, 
				 Connection.RequestHandler rh, String cached) {
	if (cached != null) {
	    cached = cached.trim(); 
	    if (cached.startsWith ("max-age=0")) {
		return setupRevalidation (con, req, rh);
	    } else {
		Date now = new Date ();
		NCacheEntry entry = rh.entry;
		if (checkMaxAge (cached, "max-age=", entry.getCacheTime (), now) ||
		    checkMaxAge (cached, "s-maxage=", entry.getCacheTime (), now)) {
		    con.setMayUseCache (false);
		    return false;
		}
	    }
	}
	return false;
    }
    
    private boolean checkMaxAge (String cached, String type, 
				 long cachetime, Date now) {
	if (cached.startsWith (type)) {
	    String secs = cached.substring (type.length ());
	    try {
		long l = Long.parseLong (secs) * 1000;
		long ad = now.getTime () - cachetime;
		if (ad > l)
		    return true;
	    } catch (NumberFormatException e) {
		logger.logError (Logger.WARN, "Bad number for max-age: '" + 
				 cached.substring (8) + "'");		    
	    }
	}
	return false;
    }

    protected boolean checkMaxAge (Connection con, HTTPHeader req, 
				   Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	if (entry == null)
	    return false;
 	List ccs = req.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    if (checkMaxAge (con, req, rh, (String)ccs.get (i)))
	 	return true;
	}
	return false;
    }

    private boolean checkNoCacheHeader (List v) {
	for (int i = 0; i < v.size (); i++) {
	    String nc = (String)v.get (i);
	    if (nc.equals ("no-cache")) {
		return true;
	    }
	}
	return false;
    }

    private boolean checkNoCache (Connection con, HTTPHeader header, 
				  Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	if (entry == null)
	    return false;
	boolean noCache = false;
	// Only check the response header, 
	// request headers with no-cache == refetch.
	HTTPHeader resp = rh.getDataHook ();
	noCache = checkNoCacheHeader (resp.getHeaders ("Cache-Control"));
	if (noCache) {
	    return setupRevalidation (con, header, rh);
	}
	return false;
    }

    private boolean checkQuery (Connection con, HTTPHeader header, 
				Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	if (entry == null)
	    return false;
	String uri = header.getRequestURI ();
	int i = uri.indexOf ('?');
	if (i >= 0) {
	    return setupRevalidation (con, header, rh);
	}
	return false;
    }

    protected long getCacheControlValue (HTTPHeader header, String cc) {
	List v = header.getHeaders ("Cache-Control");
	for (int i = 0; i < v.size (); i++) {
	    String ncc = (String)v.get (i);
	    StringTokenizer st = new StringTokenizer (ncc, ",");
	    while (st.hasMoreTokens ()) {
		String nc = st.nextToken ().trim ();
		if (nc.startsWith (cc)) {
		    return Long.parseLong (nc.substring (cc.length ()));
		}
	    }
	}
	return -1;
    }

    private boolean checkMinFresh (Connection con, HTTPHeader header, 
					  Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	if (entry == null)
	    return false;
	long minFresh = getCacheControlValue (header, "min-fresh=");
	if (minFresh == -1)
	    return false;
	long maxAge = getCacheControlValue (rh.getDataHook (), "max-age=");
	if (maxAge == -1)
	    return false;
	long currentAge = (System.currentTimeMillis () - 
			   entry.getCacheTime ()) / 1000;
	if ((maxAge - currentAge) < minFresh)
	    return setupRevalidation (con, header, rh);
	return false;
    }

    private boolean checkRevalidation (Connection con, HTTPHeader header, 
				       Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	if (entry == null)
	    return false;
	
	HTTPHeader resp = rh.getDataHook ();
	List v = resp.getHeaders ("Cache-Control");
	for (int i = 0; i < v.size (); i++) {
	    String ncc = (String)v.get (i);
	    StringTokenizer st = new StringTokenizer (ncc, ",");
	    while (st.hasMoreTokens ()) {
		String nc = st.nextToken ().trim ();
		if (nc.equals ("must-revalidate") ||
		    nc.equals ("proxy-revalidate")) {
		    con.setMustRevalidate (true);
		    long maxAge = getCacheControlValue (rh.getDataHook (), "max-age=");
		    if (maxAge >= 0) {
			long currentAge = (System.currentTimeMillis () - 
					   entry.getCacheTime ()) / 1000;
			if (maxAge == 0 || currentAge > maxAge) {
			    return setupRevalidation (con, header, rh);
			}
		    }
		} else if (nc.startsWith ("s-maxage=")) {
		    con.setMustRevalidate (true);
		    long sm = Long.parseLong (nc.substring ("s-maxage=".length ()));
		    if (sm >= 0) {
			long currentAge = (System.currentTimeMillis () - 
					   entry.getCacheTime ()) / 1000;
			if (sm == 0 || currentAge > sm) {			
			    return setupRevalidation (con, header, rh);
			}
		    }
		}
	    }
	}
	return false;
    }   

    private boolean setupRevalidation (Connection con, HTTPHeader req, 
				       Connection.RequestHandler rh) {
	NCacheEntry entry = rh.entry;
	con.setMayUseCache (false);
	String method = req.getMethod ();
	if (method.equals ("GET") && entry != null) {
	    HTTPHeader resp = rh.getDataHook ();
	    String etag = resp.getHeader ("ETag");
	    String lmod = resp.getHeader ("Last-Modified");
	    if (etag != null) {
		String inm = req.getHeader ("If-None-Match");
		if (inm == null) {
		    req.setHeader ("If-None-Match", etag);
		    con.setAddedINM (true);
		}
		return true;
	    } else if (lmod != null) {
		String ims = req.getHeader ("If-Modified-Since");
		if (ims == null) {
		    req.setHeader ("If-Modified-Since", lmod);
		    con.setAddedIMS (true);
		}
		return true;
	    } else {
		con.setMayUseCache (false);
		return false;
	    }
	}
	return false;
    }
}
