package rabbit.util;

// $Id: ThreadPool.java,v 1.7 2005/08/02 20:50:21 robo Exp $ 

import java.util.List;
import java.util.ArrayList;

/** This is a thread pool. It uses RestartableThread.
 */
public class ThreadPool<T> {
    
    private List<T> free_threads;
    private List<T> used_threads;
    private RestartableThreadFactory<T> tf;
    private int limit = Integer.MAX_VALUE;

    /** Create a ThreadPool.
     * @param tf the RestartableThreadFactory that is used as more Threads are needed.
     * @param limit dont create more threads than this.
     */
    public ThreadPool (RestartableThreadFactory<T> tf, int limit) {
	this.tf = tf;
	this.limit = limit;
	free_threads = new ArrayList<T> ();
	used_threads = new ArrayList<T> ();
    }
    
    /** Set the limit.
     * @param limit the new limit of concurrent threads.
     */
    public void setLimit (int limit) {
	this.limit = limit;
    }

    /** Get the current limit of concurrent threads.
     * @return the limit of threads.
     */
    public int getLimit () {
	return limit;
    }

    /** Get a Thread. This will wait for a thread 
     *  to be returned if the limit is currently reached.
     */
    public synchronized T getThread () {
	while (free_threads.size () == 0 && 
	       used_threads.size () == limit) {
	    try {
		wait ();   // wait for a thread to become free.
	    } catch (InterruptedException e) {
		// ignore.
	    }
	}
	
	if (free_threads.size () == 0) {
	    // we know that used_threads.size () < limit now...
	    T t = tf.createThread ();
	    used_threads.add (t);
	    return t;
	} else {
	    T t = free_threads.get (free_threads.size () - 1);
	    free_threads.remove (free_threads.size () - 1);
	    used_threads.add (t);
	    return t;
	}
    }

    /** Return a thread to the pool.
     *  If the thread was never part of this pool it will be ignored.
     * @param t the thread to return.
     */
    public synchronized void returnThread (T t) {
	int i = used_threads.indexOf (t);
	if (i == -1) {
	    System.err.println ("Unknown thread: " + t);
	    return;
	}
	boolean shouldnotify = ((free_threads.size () == 0) && 
				(used_threads.size () == limit));
	used_threads.remove (i);
	free_threads.add (t);
	if (shouldnotify)
	    notify ();
    }

    /** Set the factory to use for creating threads.
     * @param tf the new factory.
     */
    public void setFactory (RestartableThreadFactory<T> tf) {
	this.tf = tf;
    }
    
    /** Get the current factory.
     * @return the current factory.
     */
    public RestartableThreadFactory getFactory () {
	return tf;
    }

    /** The current usage of the threadpool */
    public static class Usage {
	private int freeSize;
	private int usedSize;
	
	public void setUsage (int freeSize, int usedSize) {
	    this.freeSize = freeSize;
	    this.usedSize = usedSize;
	}

	public int getFreeSize () {
	    return freeSize;
	}
	
	public int getUsedSize () {
	    return usedSize;
	}
    }
    
    /** Get the current usage of this thread pool.
     */
    public Usage getUsage () {
	Usage p = new Usage ();
	synchronized (this) {
	    p.setUsage (free_threads.size (), used_threads.size ());
	}
	return p;
    }
}
